// Copyright HTC Corporation. All Rights Reserved.

#include "WaveOpenXRSettings.h"

DEFINE_LOG_CATEGORY_STATIC(WaveOpenXRSettings, Display, All);

UWaveOpenXRSettings::UWaveOpenXRSettings(const FObjectInitializer& ObjectInitializer) :
	Super(ObjectInitializer),
	bHandInteractionEnable(false),
	bWristTrackerEnable(false),
	bControllerFocus3Enable(true)
{
}

#if WITH_EDITOR

static FName DeveloperSettingsConsoleVariableMetaFName(TEXT("ConsoleVariable"));

static void ImportConsoleVariableValues(UWaveOpenXRSettings * settings)
{
	for (FProperty* Property = settings->GetClass()->PropertyLink; Property; Property = Property->PropertyLinkNext)
	{
		if (!Property->HasAnyPropertyFlags(CPF_Config))
		{
			continue;
		}

		const FString& CVarName = Property->GetMetaData(DeveloperSettingsConsoleVariableMetaFName);
		if (!CVarName.IsEmpty())
		{
			IConsoleVariable* CVar = IConsoleManager::Get().FindConsoleVariable(*CVarName);
			if (CVar)
			{
				if (Property->ImportText(*CVar->GetString(), Property->ContainerPtrToValuePtr<uint8>(settings, 0), PPF_ConsoleVariable, settings) == NULL)
				{
					UE_LOG(WaveOpenXRSettings, Error, TEXT("%s import failed for %s on console variable %s (=%s)"), *settings->GetClass()->GetName(), *Property->GetName(), *CVarName, *CVar->GetString());
				}
			}
			else
			{
				UE_LOG(WaveOpenXRSettings, Fatal, TEXT("%s failed to find console variable %s for %s"), *settings->GetClass()->GetName(), *CVarName, *Property->GetName());
			}
		}
	}
}

static void ExportValuesToConsoleVariables(UWaveOpenXRSettings * settings, FProperty* PropertyThatChanged)
{
	if (PropertyThatChanged)
	{
		const FString& CVarName = PropertyThatChanged->GetMetaData(DeveloperSettingsConsoleVariableMetaFName);
		if (!CVarName.IsEmpty())
		{
			IConsoleVariable* CVar = IConsoleManager::Get().FindConsoleVariable(*CVarName);
			if (CVar && (CVar->GetFlags() & ECVF_ReadOnly) == 0)
			{
				FByteProperty* ByteProperty = CastField<FByteProperty>(PropertyThatChanged);
				if (ByteProperty != NULL && ByteProperty->Enum != NULL)
				{
					CVar->Set(ByteProperty->GetPropertyValue_InContainer(settings), ECVF_SetByProjectSetting);
				}
				else if (FEnumProperty* EnumProperty = CastField<FEnumProperty>(PropertyThatChanged))
				{
					FNumericProperty* UnderlyingProp = EnumProperty->GetUnderlyingProperty();
					void* PropertyAddress = EnumProperty->ContainerPtrToValuePtr<void>(settings);
					CVar->Set((int32)UnderlyingProp->GetSignedIntPropertyValue(PropertyAddress), ECVF_SetByProjectSetting);
				}
				else if (FBoolProperty* BoolProperty = CastField<FBoolProperty>(PropertyThatChanged))
				{
					CVar->Set((int32)BoolProperty->GetPropertyValue_InContainer(settings), ECVF_SetByProjectSetting);
				}
				else if (FIntProperty* IntProperty = CastField<FIntProperty>(PropertyThatChanged))
				{
					CVar->Set(IntProperty->GetPropertyValue_InContainer(settings), ECVF_SetByProjectSetting);
				}
				else if (FFloatProperty* FloatProperty = CastField<FFloatProperty>(PropertyThatChanged))
				{
					CVar->Set(FloatProperty->GetPropertyValue_InContainer(settings), ECVF_SetByProjectSetting);
				}
				else if (FStrProperty* StringProperty = CastField<FStrProperty>(PropertyThatChanged))
				{
					CVar->Set(*StringProperty->GetPropertyValue_InContainer(settings), ECVF_SetByProjectSetting);
				}
				else if (FNameProperty* NameProperty = CastField<FNameProperty>(PropertyThatChanged))
				{
					CVar->Set(*NameProperty->GetPropertyValue_InContainer(settings).ToString(), ECVF_SetByProjectSetting);
				}

			}
			else
			{
				UE_LOG(WaveOpenXRSettings, Warning, TEXT("CVar named '%s' marked up in %s was not found or is set to read-only"), *CVarName, *settings->GetClass()->GetName());
			}
		}
	}
}

void UWaveOpenXRSettings::PostEditChangeProperty(FPropertyChangedEvent& PropertyChangedEvent)
{
	Super::PostEditChangeProperty(PropertyChangedEvent);

	if (PropertyChangedEvent.Property)
	{
		ExportValuesToConsoleVariables(this, PropertyChangedEvent.Property);
	}

	//Update and write back to DefaultEngine.ini immediately.
	TryUpdateDefaultConfigFile();
}
#endif

void UWaveOpenXRSettings::PostInitProperties()
{
	Super::PostInitProperties();
#if WITH_EDITOR
        if (IsTemplate())
        {
                ImportConsoleVariableValues(this);
        }
#endif // #if WITH_EDITOR
}
