// Copyright HTC Corporation. All Rights Reserved.

#include "WaveOpenXRFocus3Controller.h"

#include "InputCoreTypes.h"
#include "WaveOpenXRExtensionPluginDefinition.h"

#define LOCTEXT_NAMESPACE "FWaveOpenXRFocus3ControllerModule"

DEFINE_LOG_CATEGORY(LogWaveOpenXRFocus3Controller);

namespace Focus3Keys
{
	/**
	 * Naming rule: (upper case will be considered as lower case)
	 * Identifier (3nd word):
	 *		"grip" => <hand>/input/squeeze
	 * Component (4th word):
	 *		"axis" => <hand>/input/<Identifier>/value
	 *		"click" => <hand>/input/<Identifier>/click //if (Identifier != "trigger" && Identifier != "squeeze")
	 *		"up", "down", "left", "right" => n/a
	 *
	 * Limitation:
	 * 1. According to the naming rules, we don't support trigger/click and grip/click
	 *
	 * Issue:
	 * - If we provides trigger/click or grip/click, all input will be unavailable.
	 **/

	// "/input/squeeze/click"
	//const FKey Focus3_Left_Grip_Click("Focus3_Left_Grip_Click");
	//const FKey Focus3_Right_Grip_Click("Focus3_Right_Grip_Click");
	// "/input/squeeze/touch"
	const FKey Focus3_Left_Grip_Touch("Focus3_Left_Grip_Touch");
	const FKey Focus3_Right_Grip_Touch("Focus3_Right_Grip_Touch");
	// "/input/squeeze/value"
	const FKey Focus3_Left_Grip_Axis("Focus3_Left_Grip_Axis");
	const FKey Focus3_Right_Grip_Axis("Focus3_Right_Grip_Axis");

	// "/input/trigger/click"
	//const FKey Focus3_Left_Trigger_Click("Focus3_Left_Trigger_Click");
	//const FKey Focus3_Right_Trigger_Click("Focus3_Right_Trigger_Click");
	// "/input/trigger/touch"
	const FKey Focus3_Left_Trigger_Touch("Focus3_Left_Trigger_Touch");
	const FKey Focus3_Right_Trigger_Touch("Focus3_Right_Trigger_Touch");
	// "/input/trigger/value"
	const FKey Focus3_Left_Trigger_Axis("Focus3_Left_Trigger_Axis");
	const FKey Focus3_Right_Trigger_Axis("Focus3_Right_Trigger_Axis");

	// "/input/thumbstick/click"
	const FKey Focus3_Left_Thumbstick_Click("Focus3_Left_Thumbstick_Click");
	const FKey Focus3_Right_Thumbstick_Click("Focus3_Right_Thumbstick_Click");
	// "/input/thumbstick/touch"
	const FKey Focus3_Left_Thumbstick_Touch("Focus3_Left_Thumbstick_Touch");
	const FKey Focus3_Right_Thumbstick_Touch("Focus3_Right_Thumbstick_Touch");
	// "/input/thumbstick/x"
	const FKey Focus3_Left_Thumbstick_X("Focus3_Left_Thumbstick_X");
	const FKey Focus3_Right_Thumbstick_X("Focus3_Right_Thumbstick_X");
	// "/input/thumbstick/y"
	const FKey Focus3_Left_Thumbstick_Y("Focus3_Left_Thumbstick_Y");
	const FKey Focus3_Right_Thumbstick_Y("Focus3_Right_Thumbstick_Y");

	// "/input/thumbrest/touch"
	const FKey Focus3_Left_Thumbrest_Touch("Focus3_Left_Thumbrest_Touch");
	const FKey Focus3_Right_Thumbrest_Touch("Focus3_Right_Thumbrest_Touch");

	/// ---- Left hand only ----
	// "/input/menu/click"
	const FKey Focus3_Left_Menu_Click("Focus3_Left_Menu_Click");
	// "/input/x/click"
	const FKey Focus3_Left_X_Click("Focus3_Left_X_Click");
	// "/input/y/click"
	const FKey Focus3_Left_Y_Click("Focus3_Left_Y_Click");

	/// ---- Right hand only ----
	// "/input/a/click"
	const FKey Focus3_Right_A_Click("Focus3_Right_A_Click");
	// "/input/b/click"
	const FKey Focus3_Right_B_Click("Focus3_Right_B_Click");
}

void FWaveOpenXRFocus3ControllerModule::StartupModule()
{
	CHECK_WAVEOPENXR_EXTENSION_FEATURE(wave.ControllerFocus3.enable);
	RegisterOpenXRExtensionModularFeature();

#pragma region Focus3 Keys
	EKeys::AddMenuCategoryDisplayInfo("Focus3", LOCTEXT("Focus3SubCategory", "Vive Focus3"), TEXT("GraphEditor.PadEvent_16x"));

	// "/input/squeeze/click"
	//EKeys::AddKey(FKeyDetails(Focus3Keys::Focus3_Left_Grip_Click, LOCTEXT("Focus3_Left_Grip_Click", "Focus3 (L) Grip"), FKeyDetails::GamepadKey | FKeyDetails::NotBlueprintBindableKey, "Focus3"));
	//EKeys::AddKey(FKeyDetails(Focus3Keys::Focus3_Right_Grip_Click, LOCTEXT("Focus3_Right_Grip_Click", "Focus3 (R) Grip"), FKeyDetails::GamepadKey | FKeyDetails::NotBlueprintBindableKey, "Focus3"));
	// "/input/squeeze/touch"
	EKeys::AddKey(FKeyDetails(Focus3Keys::Focus3_Left_Grip_Touch, LOCTEXT("Focus3_Left_Grip_Touch", "Focus3 (L) Grip Touch"), FKeyDetails::GamepadKey | FKeyDetails::NotBlueprintBindableKey, "Focus3"));
	EKeys::AddKey(FKeyDetails(Focus3Keys::Focus3_Right_Grip_Touch, LOCTEXT("Focus3_Right_Grip_Touch", "Focus3 (R) Grip Touch"), FKeyDetails::GamepadKey | FKeyDetails::NotBlueprintBindableKey, "Focus3"));
	// "/input/squeeze/value"
	EKeys::AddKey(FKeyDetails(Focus3Keys::Focus3_Left_Grip_Axis, LOCTEXT("Focus3_Left_Grip_Axis", "Focus3 (L) Grip Axis"), FKeyDetails::GamepadKey | FKeyDetails::Axis1D | FKeyDetails::NotBlueprintBindableKey, "Focus3"));
	EKeys::AddKey(FKeyDetails(Focus3Keys::Focus3_Right_Grip_Axis, LOCTEXT("Focus3_Right_Grip_Axis", "Focus3 (R) Grip Axis"), FKeyDetails::GamepadKey | FKeyDetails::Axis1D | FKeyDetails::NotBlueprintBindableKey, "Focus3"));

	// "/input/trigger/click"
	//EKeys::AddKey(FKeyDetails(Focus3Keys::Focus3_Left_Trigger_Click, LOCTEXT("Focus3_Left_Trigger_Click", "Focus3 (L) Trigger"), FKeyDetails::GamepadKey | FKeyDetails::NotBlueprintBindableKey, "Focus3"));
	//EKeys::AddKey(FKeyDetails(Focus3Keys::Focus3_Right_Trigger_Click, LOCTEXT("Focus3_Right_Trigger_Click", "Focus3 (R) Trigger"), FKeyDetails::GamepadKey | FKeyDetails::NotBlueprintBindableKey, "Focus3"));
	// "/input/trigger/touch"
	EKeys::AddKey(FKeyDetails(Focus3Keys::Focus3_Left_Trigger_Touch, LOCTEXT("Focus3_Left_Trigger_Touch", "Focus3 (L) Trigger Touch"), FKeyDetails::GamepadKey | FKeyDetails::NotBlueprintBindableKey, "Focus3"));
	EKeys::AddKey(FKeyDetails(Focus3Keys::Focus3_Right_Trigger_Touch, LOCTEXT("Focus3_Right_Trigger_Touch", "Focus3 (R) Trigger Touch"), FKeyDetails::GamepadKey | FKeyDetails::NotBlueprintBindableKey, "Focus3"));
	// "/input/trigger/value"
	EKeys::AddKey(FKeyDetails(Focus3Keys::Focus3_Left_Trigger_Axis, LOCTEXT("Focus3_Left_Trigger_Axis", "Focus3 (L) Trigger Axis"), FKeyDetails::GamepadKey | FKeyDetails::Axis1D | FKeyDetails::NotBlueprintBindableKey, "Focus3"));
	EKeys::AddKey(FKeyDetails(Focus3Keys::Focus3_Right_Trigger_Axis, LOCTEXT("Focus3_Right_Trigger_Axis", "Focus3 (R) Trigger Axis"), FKeyDetails::GamepadKey | FKeyDetails::Axis1D | FKeyDetails::NotBlueprintBindableKey, "Focus3"));

	// "/input/thumbstick/click"
	EKeys::AddKey(FKeyDetails(Focus3Keys::Focus3_Left_Thumbstick_Click, LOCTEXT("Focus3_Left_Thumbstick_Click", "Focus3 (L) Thumbstick"), FKeyDetails::GamepadKey | FKeyDetails::NotBlueprintBindableKey, "Focus3"));
	EKeys::AddKey(FKeyDetails(Focus3Keys::Focus3_Right_Thumbstick_Click, LOCTEXT("Focus3_Right_Thumbstick_Click", "Focus3 (R) Thumbstick"), FKeyDetails::GamepadKey | FKeyDetails::NotBlueprintBindableKey, "Focus3"));
	// "/input/thumbstick/touch"
	EKeys::AddKey(FKeyDetails(Focus3Keys::Focus3_Left_Thumbstick_Touch, LOCTEXT("Focus3_Left_Thumbstick_Touch", "Focus3 (L) Thumbstick Touch"), FKeyDetails::GamepadKey | FKeyDetails::NotBlueprintBindableKey, "Focus3"));
	EKeys::AddKey(FKeyDetails(Focus3Keys::Focus3_Right_Thumbstick_Touch, LOCTEXT("Focus3_Right_Thumbstick_Touch", "Focus3 (R) Thumbstick Touch"), FKeyDetails::GamepadKey | FKeyDetails::NotBlueprintBindableKey, "Focus3"));
	// "/input/thumbstick/x"
	EKeys::AddKey(FKeyDetails(Focus3Keys::Focus3_Left_Thumbstick_X, LOCTEXT("Focus3_Left_Thumbstick_X", "Focus3 (L) Thumbstick X"), FKeyDetails::GamepadKey | FKeyDetails::Axis1D | FKeyDetails::NotBlueprintBindableKey, "Focus3"));
	EKeys::AddKey(FKeyDetails(Focus3Keys::Focus3_Right_Thumbstick_X, LOCTEXT("Focus3_Right_Thumbstick_X", "Focus3 (R) Thumbstick X"), FKeyDetails::GamepadKey | FKeyDetails::Axis1D | FKeyDetails::NotBlueprintBindableKey, "Focus3"));
	// "/input/thumbstick/y"
	EKeys::AddKey(FKeyDetails(Focus3Keys::Focus3_Left_Thumbstick_Y, LOCTEXT("Focus3_Left_Thumbstick_Y", "Focus3 (L) Thumbstick Y"), FKeyDetails::GamepadKey | FKeyDetails::Axis1D | FKeyDetails::NotBlueprintBindableKey, "Focus3"));
	EKeys::AddKey(FKeyDetails(Focus3Keys::Focus3_Right_Thumbstick_Y, LOCTEXT("Focus3_Right_Thumbstick_Y", "Focus3 (R) Thumbstick Y"), FKeyDetails::GamepadKey | FKeyDetails::Axis1D | FKeyDetails::NotBlueprintBindableKey, "Focus3"));

	// "/input/thumbrest/touch"
	EKeys::AddKey(FKeyDetails(Focus3Keys::Focus3_Left_Thumbrest_Touch, LOCTEXT("Focus3_Left_Thumbrest_Touch", "Focus3 (L) Thumbrest Touch"), FKeyDetails::GamepadKey | FKeyDetails::NotBlueprintBindableKey, "Focus3"));
	EKeys::AddKey(FKeyDetails(Focus3Keys::Focus3_Right_Thumbrest_Touch, LOCTEXT("Focus3_Right_Thumbrest_Touch", "Focus3 (R) Thumbrest Touch"), FKeyDetails::GamepadKey | FKeyDetails::NotBlueprintBindableKey, "Focus3"));

	/// ---- Left hand only ----
	// "/input/menu/click"
	EKeys::AddKey(FKeyDetails(Focus3Keys::Focus3_Left_Menu_Click, LOCTEXT("Focus3_Left_Menu_Click", "Focus3 (L) Menu"), FKeyDetails::GamepadKey | FKeyDetails::NotBlueprintBindableKey, "Focus3"));
	// "/input/x/click"
	EKeys::AddKey(FKeyDetails(Focus3Keys::Focus3_Left_X_Click, LOCTEXT("Focus3_Left_X_Click", "Focus3 (L) X Press"), FKeyDetails::GamepadKey | FKeyDetails::NotBlueprintBindableKey, "Focus3"));
	// "/input/y/click"
	EKeys::AddKey(FKeyDetails(Focus3Keys::Focus3_Left_Y_Click, LOCTEXT("Focus3_Left_Y_Click", "Focus3 (L) Y Press"), FKeyDetails::GamepadKey | FKeyDetails::NotBlueprintBindableKey, "Focus3"));

	/// ---- Right hand only ----
	// "/input/a/click"
	EKeys::AddKey(FKeyDetails(Focus3Keys::Focus3_Right_A_Click, LOCTEXT("Focus3_Right_A_Click", "Focus3 (R) A Press"), FKeyDetails::GamepadKey | FKeyDetails::NotBlueprintBindableKey, "Focus3"));
	// "/input/b/click"
	EKeys::AddKey(FKeyDetails(Focus3Keys::Focus3_Right_B_Click, LOCTEXT("Focus3_Right_B_Click", "Focus3 (R) B Press"), FKeyDetails::GamepadKey | FKeyDetails::NotBlueprintBindableKey, "Focus3"));
#pragma endregion
}

void FWaveOpenXRFocus3ControllerModule::ShutdownModule()
{
	IModularFeatures::Get().UnregisterModularFeature(GetModularFeatureName(), this);
}


bool FWaveOpenXRFocus3ControllerModule::GetRequiredExtensions(TArray<const ANSICHAR*>& OutExtensions)
{
	OutExtensions.Add("XR_HTC_vive_focus3_controller_interaction");
	return true;
}

bool FWaveOpenXRFocus3ControllerModule::GetInteractionProfile(XrInstance InInstance, FString& OutKeyPrefix, XrPath& OutPath, bool& OutHasHaptics)
{
	OutKeyPrefix = "Focus3";
	OutHasHaptics = true;
	return xrStringToPath(InInstance, "/interaction_profiles/htc/vive_focus3_controller", &OutPath) == XR_SUCCESS;
}

#undef LOCTEXT_NAMESPACE
	
IMPLEMENT_MODULE(FWaveOpenXRFocus3ControllerModule, WaveOpenXRFocus3Controller)
