// Copyright HTC Corporation. All Rights Reserved.

#include "WaveOpenXRGazePointer.h"
#include "WaveGazeInteractInterface.h"

#include "CollisionQueryParams.h"
#include "Components/WidgetComponent.h"
#include "DrawDebugHelpers.h"
#include "Engine/Blueprint.h"
#include "Engine/Engine.h"
#include "Engine/World.h"
#include "Kismet/GameplayStatics.h"
#include "Kismet/KismetMathLibrary.h"
#include "Kismet/KismetSystemLibrary.h"
#include "Math/UnrealMathUtility.h"
#include "EyeTrackerFunctionLibrary.h"

DEFINE_LOG_CATEGORY_STATIC(LogWaveOpenXRGazePointer, Log, All);

UWaveOpenXRGazePointer::UWaveOpenXRGazePointer()
	: SpotPointer(nullptr)
	, RingPointer(nullptr)
	, RollingRingPointer(nullptr)
	, WidgetInteraction(nullptr)
	, spotPointerMesh(nullptr)
	, spotPointerMaterial(nullptr)
	, ringPointerMesh(nullptr)
	, ringPointerMaterial(nullptr)
	, rollingRingPointerMesh(nullptr)
	, rollingRingPointerMaterial(nullptr)
	, worldLocation(FVector::ZeroVector)
	, worldRotation(FRotator::ZeroRotator)
	, tracedPhysics(false)
	, hit(ForceInit)
	, focusActor(nullptr)
	, tracedWidget(false)
	, tracedWidgetPosition(FVector::ZeroVector)
	, gazeOnTime(0)
	, pressState(false)
{
	// Set this component to be initialized when the game starts, and to be ticked every frame.  You can turn these features
	// off to improve performance if you don't need them.
	PrimaryComponentTick.bCanEverTick = true;

	m_GazeTimer = GazeTimer;
	m_RightClickButtons = RightClickButtons;
	m_LeftClickButtons = LeftClickButtons;

	FString reticle_mesh = TEXT("StaticMesh'/WaveOpenXR/Shapes/Reticle_Mesh.Reticle_Mesh'");
	spotPointerMesh = Cast<UStaticMesh>(StaticLoadObject(UStaticMesh::StaticClass(), NULL, *reticle_mesh));
	FString m_reticle = TEXT("Material'/WaveOpenXR/Materials/M_Reticle.M_Reticle'");
	spotPointerMaterial = Cast<UMaterialInterface>(StaticLoadObject(UMaterialInterface::StaticClass(), NULL, *m_reticle));
	SpotPointer = CreateDefaultSubobject<UStaticMeshComponent>(TEXT("SpotPointer"));
	//SpotPointer = NewObject<UStaticMeshComponent>(this, TEXT("SpotPointer"));
	SpotPointer->SetupAttachment(this);

	FString static_circle = TEXT("StaticMesh'/WaveOpenXR/Shapes/StaticCircle.StaticCircle'");
	ringPointerMesh = Cast<UStaticMesh>(StaticLoadObject(UStaticMesh::StaticClass(), NULL, *static_circle));
	FString m_staticcircle = TEXT("Material'/WaveOpenXR/Materials/M_Circle.M_Circle'");
	ringPointerMaterial = Cast<UMaterialInterface>(StaticLoadObject(UMaterialInterface::StaticClass(), NULL, *m_staticcircle));
	RingPointer = CreateDefaultSubobject<UStaticMeshComponent>(TEXT("RingPointer"));
	//RingPointer = NewObject<UStaticMeshComponent>(this, TEXT("RingPointer"));
	RingPointer->SetupAttachment(this);

	FString rolling_circle = TEXT("StaticMesh'/WaveOpenXR/Shapes/SpinningCircle.SpinningCircle'");
	rollingRingPointerMesh = Cast<UStaticMesh>(StaticLoadObject(UStaticMesh::StaticClass(), NULL, *rolling_circle));
	FString m_rollingcircle = TEXT("Material'/WaveOpenXR/Materials/M_RollingCircle.M_RollingCircle'");
	rollingRingPointerMaterial = Cast<UMaterialInterface>(StaticLoadObject(UMaterialInterface::StaticClass(), NULL, *m_rollingcircle));
	RollingRingPointer = CreateDefaultSubobject<UStaticMeshComponent>(TEXT("RollingRingPointer"));
	//RollingRingPointer = NewObject<UStaticMeshComponent>(this, TEXT("RollingRingPointer"));
	RollingRingPointer->SetupAttachment(this);

	WidgetInteraction = CreateDefaultSubobject<UWidgetInteractionComponent>(TEXT("WidgetInteraction"));
	//WidgetInteraction = NewObject<UWidgetInteractionComponent>(this, TEXT("WidgetInteraction"));
	WidgetInteraction->SetupAttachment(this);
}

void UWaveOpenXRGazePointer::CheckLogInterval()
{
	logFrame++;
	logFrame = logFrame % klogFrameCount;
	if (logFrame == 0)
		printIntervalLog = true;
	else
		printIntervalLog = false;
}

#pragma region
void UWaveOpenXRGazePointer::SetWidgetPointerIndex(int32 pointerIndex)
{
	if (WidgetInteraction)
	{
		WidgetPointerIndex = pointerIndex;
		WidgetInteraction->PointerIndex = WidgetPointerIndex;
		UE_LOG(LogWaveOpenXRGazePointer, Log, TEXT("SetWidgetPointerIndex() %f"), WidgetInteraction->PointerIndex);
	}
}
void UWaveOpenXRGazePointer::SetTimerEnable(bool enable)
{
	TimerControl = enable;
	UE_LOG(LogWaveOpenXRGazePointer, Log, TEXT("SetTimerEnable() %d"), (uint8_t)TimerControl);
}
void UWaveOpenXRGazePointer::SetupGazeTimer(float duration)
{
	GazeTimer = duration;
	UE_LOG(LogWaveOpenXRGazePointer, Log, TEXT("SetupGazeTimer() %f"), GazeTimer);
}
void UWaveOpenXRGazePointer::SetButtonEnable(bool enable)
{
	ButtonControl = enable;
	UE_LOG(LogWaveOpenXRGazePointer, Log, TEXT("SetButtonEnable() %d"), (uint8_t)ButtonControl);
}
void UWaveOpenXRGazePointer::SetupGazeButton(const TArray<FKey> leftButtons, const TArray<FKey> rightButtons)
{
	if (rightButtons.Num() > 0)
		RightClickButtons = rightButtons;
	if (leftButtons.Num() > 0)
		LeftClickButtons = leftButtons;
}
#pragma endregion Public APIs

void UWaveOpenXRGazePointer::UpdateCustomizedVariables()
{
	if (m_RightClickButtons != RightClickButtons || m_LeftClickButtons != LeftClickButtons)
	{
		m_RightClickButtons = RightClickButtons;
		m_LeftClickButtons = LeftClickButtons;

		for (int i = 0; i < m_RightClickButtons.Num(); i++)
			UE_LOG(LogWaveOpenXRGazePointer, Log, TEXT("UpdateCustomizedVariables() ButtonControl %d, m_RightClickButtons[%d]=%s"), (uint8)ButtonControl, i, *m_RightClickButtons[i].ToString());
		for (int i = 0; i < m_LeftClickButtons.Num(); i++)
			UE_LOG(LogWaveOpenXRGazePointer, Log, TEXT("UpdateCustomizedVariables() ButtonControl %d, m_LeftClickButtons[%d]=%s"), (uint8)ButtonControl, i, *m_LeftClickButtons[i].ToString());
	}

	if (m_GazeTimer != GazeTimer)
	{
		m_GazeTimer = GazeTimer;
		m_GazeTimer = FMath::Clamp(m_GazeTimer, kGazeTimerMin, kGazeTimerMax);
		UE_LOG(LogWaveOpenXRGazePointer, Log, TEXT("UpdateCustomizedVariables() TimerControl %d,  m_GazeTimer %f"), (uint8)TimerControl, m_GazeTimer);
	}
}

void UWaveOpenXRGazePointer::BeginPlay()
{
	Super::BeginPlay();

	ConfigureResources();
}
void UWaveOpenXRGazePointer::TickComponent(float DeltaTime, ELevelTick TickType, FActorComponentTickFunction* ThisTickFunction)
{
	Super::TickComponent(DeltaTime, TickType, ThisTickFunction);

	CheckLogInterval();

	UpdateCustomizedVariables();

	if (!IsGazeInteractable())
		return;

	worldLocation = GetComponentLocation();// GetOwner()->GetActorLocation();
	worldRotation = GetEyeRotation();// GetComponentRotation();
	FVector end_pos = (GetEyeVector() * TraceDistance) + worldLocation;

	bool clicked = IsClicked();

	FCollisionQueryParams CollisionParams;
	SetIgnoreActors(CollisionParams);
	tracedPhysics = GetWorld()->LineTraceSingleByChannel(hit, worldLocation, end_pos, ECC_Camera, CollisionParams);
	//DrawDebugLine(GetWorld(), worldLocation, end_pos, FColor::Green, false, 0.5f);
	if (tracedPhysics)
	{
		AActor* hit_actor = hit.GetActor();
		//GEngine->AddOnScreenDebugMessage(-1, 1.f, FColor::Red, FString::Printf(TEXT("Trace hit: %s"), *hit.GetActor()->GetName()));
		//LOGD(LogWaveOpenXRGazePointer, "TickComponent() Hitting: %s (%s)", *hit_actor->GetName(), *hit.ImpactPoint.ToString());

		if (focusActor != hit_actor) // Enter & Exit
		{
			if (focusActor && (focusActor->GetClass()->ImplementsInterface(UWaveGazeInteractInterface::StaticClass())))
			{
				// Exit previous actor
				IWaveGazeInteractInterface::Execute_OnGazeExit(focusActor, GetOwner());
				UE_LOG(LogWaveOpenXRGazePointer, Log, TEXT("TickComponent() Exit: %s"), *focusActor->GetName());
			}
			if (hit_actor && (hit_actor->GetClass()->ImplementsInterface(UWaveGazeInteractInterface::StaticClass())))
			{
				// Enter current actor
				IWaveGazeInteractInterface::Execute_OnGazeEnter(hit_actor, GetOwner());
				UE_LOG(LogWaveOpenXRGazePointer, Log, TEXT("TickComponent() Enter: %s"), *hit_actor->GetName());

				// Reset the timer when switching the physical gaze target.
				gazeOnTime = UGameplayStatics::GetTimeSeconds(GetWorld());
			}

			focusActor = hit_actor;
		}
		else // Hover & Click
		{
			if (focusActor && (focusActor->GetClass()->ImplementsInterface(UWaveGazeInteractInterface::StaticClass())))
			{
				IWaveGazeInteractInterface::Execute_OnGazeHover(focusActor, GetOwner());
				if (clicked || IsTimeout())
				{
					IWaveGazeInteractInterface::Execute_OnGazeClick(focusActor, GetOwner());
					UE_LOG(LogWaveOpenXRGazePointer, Log, TEXT("TickComponent() Click: %s"), *hit_actor->GetName());

					// Reset the gaze timer on click.
					gazeOnTime = UGameplayStatics::GetTimeSeconds(GetWorld());
				}
			}
		}
	}
	else
	{
		if (focusActor)
		{
			if (focusActor->GetClass()->ImplementsInterface(UWaveGazeInteractInterface::StaticClass()))
			{
				// Exit previous actor
				IWaveGazeInteractInterface::Execute_OnGazeExit(focusActor, GetOwner());
				UE_LOG(LogWaveOpenXRGazePointer, Log, TEXT("TickComponent() Exit: %s"), *focusActor->GetName());
			}
			focusActor = nullptr;
		}

		if (WidgetInteraction) // Click
		{
			WidgetInteraction->SetWorldLocationAndRotation(worldLocation, worldRotation);
			tracedWidget = WidgetInteraction->IsOverFocusableWidget();
			if (tracedWidget)
			{
				UWidgetComponent* widget = WidgetInteraction->GetHoveredWidgetComponent();
				UUserWidget* user_widget = widget->GetUserWidgetObject();
				tracedWidgetPosition = (static_cast<USceneComponent *>(widget))->GetComponentLocation();
				if (clicked || IsTimeout())
				{
					WidgetInteraction->PressPointerKey(FKey("LeftMouseButton"));
					WidgetInteraction->ReleasePointerKey(FKey("LeftMouseButton"));
					UE_LOG(LogWaveOpenXRGazePointer, Log, TEXT("TickComponent() Click widget: %s"), *user_widget->GetFullName());

					// Reset the gaze timer on click.
					gazeOnTime = UGameplayStatics::GetTimeSeconds(GetWorld());
				}
			}
			else
			{
				// Reset the timer when not gazing on any widget.
				gazeOnTime = UGameplayStatics::GetTimeSeconds(GetWorld());
			}
		}
		else
		{
			tracedWidget = false;
		}
	}

	ActivatePointer();
}

void UWaveOpenXRGazePointer::ConfigureResources()
{
	if (WidgetInteraction)
	{
		WidgetInteraction->VirtualUserIndex = WidgetVirtualUserIndex;
		WidgetInteraction->PointerIndex = WidgetPointerIndex;
		WidgetInteraction->TraceChannel = ECollisionChannel::ECC_Visibility;
		WidgetInteraction->InteractionDistance = TraceDistance;
		WidgetInteraction->InteractionSource = EWidgetInteractionSource::World;
		//WidgetInteraction->bEnableHitTesting = true;
		//WidgetInteraction->bShowDebug = true;
		//WidgetInteraction->DebugColor = FLinearColor::Red;

		UE_LOG(LogWaveOpenXRGazePointer, Log, TEXT("ConfigureResources() WidgetInteraction"));
	}

	if (SpotPointer)
	{
		if (spotPointerMesh)
		{
			SpotPointer->SetStaticMesh(spotPointerMesh);
			UE_LOG(LogWaveOpenXRGazePointer, Log, TEXT("ConfigureResources() spotPointerMesh"));
		}
		if (spotPointerMaterial)
		{
			SpotPointer->SetMaterial(0, spotPointerMaterial);
			UE_LOG(LogWaveOpenXRGazePointer, Log, TEXT("ConfigureResources() spotPointerMaterial"));
		}
		SpotPointer->SetRelativeRotation(SpotRotationOffset);
		SpotPointer->SetRelativeScale3D(SpotScale);
		SpotPointer->SetCollisionEnabled(ECollisionEnabled::Type::NoCollision);

		UE_LOG(LogWaveOpenXRGazePointer, Log, TEXT("ConfigureResources() SpotPointer"));
	}

	if (RingPointer)
	{
		if (ringPointerMesh)
		{
			RingPointer->SetStaticMesh(ringPointerMesh);
			UE_LOG(LogWaveOpenXRGazePointer, Log, TEXT("ConfigureResources() ringPointerMesh"));
		}
		if (ringPointerMaterial)
		{
			RingPointer->SetMaterial(0, ringPointerMaterial);
			UE_LOG(LogWaveOpenXRGazePointer, Log, TEXT("ConfigureResources() ringPointerMaterial"));
		}
		RingPointer->SetRelativeRotation(RingRotationOffset);
		RingPointer->SetRelativeScale3D(RingScale);
		RingPointer->SetCollisionEnabled(ECollisionEnabled::Type::NoCollision);
		RingPointer->SetHiddenInGame(true, true);

		UE_LOG(LogWaveOpenXRGazePointer, Log, TEXT("ConfigureResources() RingPointer"));
	}

	if (RollingRingPointer)
	{
		if (rollingRingPointerMesh)
		{
			RollingRingPointer->SetStaticMesh(rollingRingPointerMesh);
			UE_LOG(LogWaveOpenXRGazePointer, Log, TEXT("ConfigureResources() rollingRingPointerMesh"));
		}
		if (rollingRingPointerMaterial)
		{
			RollingRingPointer->SetMaterial(0, rollingRingPointerMaterial);
			UE_LOG(LogWaveOpenXRGazePointer, Log, TEXT("ConfigureResources() rollingRingPointerMaterial"));
		}
		RollingRingPointer->SetRelativeRotation(RingRotationOffset);
		RollingRingPointer->SetRelativeScale3D(RingScale);
		RollingRingPointer->SetCollisionEnabled(ECollisionEnabled::Type::NoCollision);
		RollingRingPointer->SetHiddenInGame(true, true);

		UE_LOG(LogWaveOpenXRGazePointer, Log, TEXT("ConfigureResources() RollingRingPointer"));
	}
}
void UWaveOpenXRGazePointer::ActivatePointer()
{
	if (SpotPointer)
	{
		SpotPointer->SetWorldLocation((GetEyeVector() * SpotDistance) + worldLocation);
		if (tracedPhysics || tracedWidget)
			SpotPointer->SetHiddenInGame(true, true);
		else
			SpotPointer->SetHiddenInGame(false, true);
	}
	if (RingPointer)
	{
		RingPointer->SetWorldLocation((GetEyeVector() * RingDistance) + worldLocation);
		RingPointer->SetWorldRotation(worldRotation + RingRotationOffset);
		if (TimerControl)
			RingPointer->SetHiddenInGame(true, true);
		else
		{
			if (tracedPhysics || tracedWidget)
				RingPointer->SetHiddenInGame(false, true);
			else
				RingPointer->SetHiddenInGame(true, true);
		}
	}
	if (RollingRingPointer)
	{
		RollingRingPointer->SetWorldLocation((GetEyeVector() * RingDistance) + worldLocation);
		RollingRingPointer->SetWorldRotation(worldRotation + RingRotationOffset);
		if (!TimerControl)
			RollingRingPointer->SetHiddenInGame(true, true);
		else
		{
			if (tracedPhysics || tracedWidget)
				RollingRingPointer->SetHiddenInGame(false, true);
			else
				RollingRingPointer->SetHiddenInGame(true, true);
		}
	}
}
bool UWaveOpenXRGazePointer::IsGazeInteractable()
{
	bool interactable = true;

	if (WidgetInteraction)
	{
		if ((interactable && !WidgetInteraction->IsActive()) ||
			(!interactable && WidgetInteraction->IsActive()))
		{
			UE_LOG(LogWaveOpenXRGazePointer, Log, TEXT("IsGazeInteractable() %d"), (uint8)interactable);
		}
		WidgetInteraction->SetActive(interactable);
	}
	if (!interactable)
	{
		if (SpotPointer)
			SpotPointer->SetHiddenInGame(true, true);
		if (RingPointer)
			RingPointer->SetHiddenInGame(true, true);
		if (RollingRingPointer)
			RollingRingPointer->SetHiddenInGame(true, true);
	}

	return interactable;
}

void UWaveOpenXRGazePointer::SetIgnoreActors(FCollisionQueryParams& collisionParams)
{
	// Ignore WaveOpenXR_LinePointer blueprint actor.
	/*FString pointer_path = TEXT("Blueprint'/WaveOpenXR/Blueprints/ControllerModel/WaveOpenXR_LinePointer.WaveOpenXR_LinePointer'");
	UBlueprint* pointer_blueprint = Cast<UBlueprint>(StaticLoadObject(UObject::StaticClass(), nullptr, *pointer_path));
	if (pointer_blueprint && pointer_blueprint->GeneratedClass->IsChildOf(AActor::StaticClass()))
	{
		TArray<AActor *> pointer_ignored;
		UGameplayStatics::GetAllActorsOfClass(GetWorld(), *pointer_blueprint->GeneratedClass, pointer_ignored);
		collisionParams.AddIgnoredActors(pointer_ignored);
		for (int i = 0; i < pointer_ignored.Num(); i++)
		{
			AActor* actor = pointer_ignored[i];
			LOGD(LogWaveOpenXRGazePointer, "SetIgnoreActors() %s", *actor->GetFullName());
		}
	}*/

	// Ignore the Pawn.
	for (const FWorldContext& Context : GEngine->GetWorldContexts())
	{
		APawn* pawn = UGameplayStatics::GetPlayerPawn(Context.World(), PlayerIndex);
		if (pawn)
		{
			collisionParams.AddIgnoredActor(pawn);
			//LOGD(LogWaveOpenXRGazePointer, "SetIgnoreActors() %s", *pawn->GetFullName());
		}
	}

	// Ignore parent actor.
	USceneComponent* parent = GetAttachParent();
	while (parent)
	{
		collisionParams.AddIgnoredActor(parent->GetOwner());
		//LOGD(LogWaveOpenXRGazePointer, "SetIgnoreActors() %s", *parent->GetOwner()->GetFullName());
		parent = parent->GetAttachParent();
	}
}

FVector UWaveOpenXRGazePointer::GetEyeVector()
{
	FVector vec = GetForwardVector();

	if (EyeTracking && UEyeTrackerFunctionLibrary::GetGazeData(m_GazeData))
	{
		vec = m_GazeData.GazeDirection;
		if (printIntervalLog)
			UE_LOG(LogWaveOpenXRGazePointer, Log, TEXT("GetEyeVector(%f, %f, %f)"), vec.X, vec.Y, vec.Z);
	}

	return vec;
}
FRotator UWaveOpenXRGazePointer::GetEyeRotation()
{
	FRotator rot = GetComponentRotation();

	if (EyeTracking && UEyeTrackerFunctionLibrary::GetGazeData(m_GazeData))
		rot = UKismetMathLibrary::MakeRotFromX(m_GazeData.GazeDirection);

	return rot;
}

#pragma region
bool UWaveOpenXRGazePointer::IsTimeout()
{
	if (!TimerControl && !ButtonControl) { TimerControl = true; }
	if (!TimerControl)
		return false;

	return UGameplayStatics::GetTimeSeconds(GetWorld()) > gazeOnTime + GazeTimer;
}
bool UWaveOpenXRGazePointer::IsClicked()
{
	if (!ButtonControl)
		return false;

	bool pressed = false;
	for (int i = 0; i < m_RightClickButtons.Num(); i++)
	{
		if (pressed) { break; }

		float keyDownTime = UGameplayStatics::GetPlayerController(GetWorld(), 0)->GetInputKeyTimeDown(m_RightClickButtons[i]);
		//UE_LOG(LogWaveOpenXRGazePointer, Log, TEXT("IsClicked() right keyDownTime %f"), keyDownTime);
		pressed = (keyDownTime > 0.2f);

		if (pressState == false && pressed == true)
			UE_LOG(LogWaveOpenXRGazePointer, Log, TEXT("IsClicked() right %s is pressed."), *(m_RightClickButtons[i].GetFName()).ToString());
	}
	for (int i = 0; i < m_LeftClickButtons.Num(); i++)
	{
		if (pressed) { break; }

		float keyDownTime = UGameplayStatics::GetPlayerController(GetWorld(), 0)->GetInputKeyTimeDown(m_LeftClickButtons[i]);
		pressed = (keyDownTime > 0.2f);

		if (pressState == false && pressed == true)
			UE_LOG(LogWaveOpenXRGazePointer, Log, TEXT("IsClicked() left %s is pressed."), *(m_LeftClickButtons[i].GetFName()).ToString());
	}

	bool clicked = (pressState == false && pressed == true);
	pressState = pressed;

	return clicked;
}
#pragma endregion Trigger Type
