// Copyright HTC Corporation. All Rights Reserved.

#pragma once

#include "CoreMinimal.h"
#include "Components/SceneComponent.h"

// Line trace and GUI
#include "GameFramework/Actor.h"
#include "Components/WidgetInteractionComponent.h"
#include "Engine/StaticMesh.h"
#include "Materials/MaterialInterface.h"
#include "Components/StaticMeshComponent.h"
#include "EyeTrackerTypes.h"
#include "InputCoreTypes.h"

// Blueprint interface
#include "WaveGazeInteractInterface.h"

#include "WaveOpenXRGazePointer.generated.h"

namespace Focus3Keys
{
	/**
	 * Naming rule: (upper case will be considered as lower case)
	 * Identifier (3nd word):
	 *		"grip" => <hand>/input/squeeze
	 * Component (4th word):
	 *		"axis" => <hand>/input/<Identifier>/value
	 *		"click" => <hand>/input/<Identifier>/click //if (Identifier != "trigger" && Identifier != "squeeze")
	 *		"up", "down", "left", "right" => n/a
	 *
	 * Limitation:
	 * 1. According to the naming rules, we don't support trigger/click and grip/click
	 *
	 * Issue:
	 * - If we provides trigger/click or grip/click, all input will be unavailable.
	 **/

	 // "/input/squeeze/click"
	 //const FKey Focus3_Left_Grip_Click("Focus3_Left_Grip_Click");
	 //const FKey Focus3_Right_Grip_Click("Focus3_Right_Grip_Click");
	 // "/input/squeeze/touch"
	const FKey Focus3_Left_Grip_Touch("Focus3_Left_Grip_Touch");
	const FKey Focus3_Right_Grip_Touch("Focus3_Right_Grip_Touch");
	// "/input/squeeze/value"
	const FKey Focus3_Left_Grip_Axis("Focus3_Left_Grip_Axis");
	const FKey Focus3_Right_Grip_Axis("Focus3_Right_Grip_Axis");

	// "/input/trigger/click"
	//const FKey Focus3_Left_Trigger_Click("Focus3_Left_Trigger_Click");
	//const FKey Focus3_Right_Trigger_Click("Focus3_Right_Trigger_Click");
	// "/input/trigger/touch"
	const FKey Focus3_Left_Trigger_Touch("Focus3_Left_Trigger_Touch");
	const FKey Focus3_Right_Trigger_Touch("Focus3_Right_Trigger_Touch");
	// "/input/trigger/value"
	const FKey Focus3_Left_Trigger_Axis("Focus3_Left_Trigger_Axis");
	const FKey Focus3_Right_Trigger_Axis("Focus3_Right_Trigger_Axis");

	// "/input/thumbstick/click"
	const FKey Focus3_Left_Thumbstick_Click("Focus3_Left_Thumbstick_Click");
	const FKey Focus3_Right_Thumbstick_Click("Focus3_Right_Thumbstick_Click");
	// "/input/thumbstick/touch"
	const FKey Focus3_Left_Thumbstick_Touch("Focus3_Left_Thumbstick_Touch");
	const FKey Focus3_Right_Thumbstick_Touch("Focus3_Right_Thumbstick_Touch");
	// "/input/thumbstick/x"
	const FKey Focus3_Left_Thumbstick_X("Focus3_Left_Thumbstick_X");
	const FKey Focus3_Right_Thumbstick_X("Focus3_Right_Thumbstick_X");
	// "/input/thumbstick/y"
	const FKey Focus3_Left_Thumbstick_Y("Focus3_Left_Thumbstick_Y");
	const FKey Focus3_Right_Thumbstick_Y("Focus3_Right_Thumbstick_Y");

	// "/input/thumbrest/touch"
	const FKey Focus3_Left_Thumbrest_Touch("Focus3_Left_Thumbrest_Touch");
	const FKey Focus3_Right_Thumbrest_Touch("Focus3_Right_Thumbrest_Touch");

	/// ---- Left hand only ----
	// "/input/menu/click"
	const FKey Focus3_Left_Menu_Click("Focus3_Left_Menu_Click");
	// "/input/x/click"
	const FKey Focus3_Left_X_Click("Focus3_Left_X_Click");
	// "/input/y/click"
	const FKey Focus3_Left_Y_Click("Focus3_Left_Y_Click");

	/// ---- Right hand only ----
	// "/input/a/click"
	const FKey Focus3_Right_A_Click("Focus3_Right_A_Click");
	// "/input/b/click"
	const FKey Focus3_Right_B_Click("Focus3_Right_B_Click");
}

UCLASS( ClassGroup=(Custom), meta=(BlueprintSpawnableComponent) )
class UWaveOpenXRGazePointer : public USceneComponent
{
	GENERATED_BODY()

public:
	UWaveOpenXRGazePointer();

protected:
	virtual void BeginPlay() override;

public:
	virtual void TickComponent(float DeltaTime, ELevelTick TickType, FActorComponentTickFunction* ThisTickFunction) override;

private:
	uint32_t logFrame = 0;
	const uint32_t klogFrameCount = 300;
	bool printIntervalLog = false;
	void CheckLogInterval();

#pragma region
protected:
	/** True to use eye tracking for gaze. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Wave|OpenXR|Gaze")
	bool EyeTracking = false;

	/** Specify the player index of your pawn. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Wave|OpenXR|Gaze")
	int32 PlayerIndex = 0;

	/** True to enable timer control for gaze. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Wave|OpenXR|Gaze")
	bool TimerControl = true;
	const float kGazeTimerMin = 1.f, kGazeTimerMax = 5.f;
	float m_GazeTimer = 2.f;
	/** Set up the timer count down duration in seconds. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Wave|OpenXR|Gaze", meta = (ClampMin = "1.0", ClampMax = "5.0"))
	float GazeTimer = 2.f;

	/** True to enable button control for gaze. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Wave|OpenXR|Gaze")
	bool ButtonControl = true;

	const TArray<FKey> kClickButtonsLeft = { Focus3Keys::Focus3_Left_Trigger_Axis };
	TArray<FKey> m_LeftClickButtons = kClickButtonsLeft;
	/** Set up the left controller buttons for click. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Wave|OpenXR|Gaze")
	TArray<FKey> LeftClickButtons = kClickButtonsLeft;

	const TArray<FKey> kClickButtonsRight = { Focus3Keys::Focus3_Right_Trigger_Axis };
	TArray<FKey> m_RightClickButtons = kClickButtonsRight;
	/** Set up the right controller buttons for click. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Wave|OpenXR|Gaze")
	TArray<FKey> RightClickButtons = kClickButtonsRight;

	/** Set up the available distance of gaze LineTrace. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category="Wave|OpenXR|Gaze")
	float TraceDistance = 10000.f;

	/** Set up the scale of the gaze spot. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Wave|OpenXR|Gaze")
	FVector SpotScale = FVector(0.015f, 0.015f, 0.015f);
	/** Set up the rotation offset of the gaze spot. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Wave|OpenXR|Gaze")
	FRotator SpotRotationOffset = FRotator(0, 0, 0);
	/** Set up the initial distance of the gaze spot. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Wave|OpenXR|Gaze")
	float SpotDistance = 100.f;

	/** Set up the scale of the gaze ring. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Wave|OpenXR|Gaze")
	FVector RingScale = FVector(0.03f, 0.03f, 0.03f);
	/** Set up the rotation offset of the gaze ring. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Wave|OpenXR|Gaze")
	FRotator RingRotationOffset = FRotator(90, 0, 0);
	/** Set up the initial distance of the gaze ring. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Wave|OpenXR|Gaze")
	float RingDistance = 100.f;

	/** Set up the VirtualUserIndex of the gaze widget. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Wave|OpenXR|Gaze")
	int32 WidgetVirtualUserIndex = 0;
	/** Set up the PointerIndex of the gaze widget. */
	UPROPERTY(EditAnywhere, BlueprintReadWrite, Category = "Wave|OpenXR|Gaze")
	int32 WidgetPointerIndex = 0;
#pragma endregion Customized Variables

public:
	void SetWidgetPointerIndex(int32 pointerIndex);
	void SetTimerEnable(bool enable);
	void SetupGazeTimer(float duration);
	void SetButtonEnable(bool enable);
	void SetupGazeButton(const TArray<FKey> leftButtons, const TArray<FKey> rightButtons);

private:
	void UpdateCustomizedVariables();

// Resources
protected:
	UStaticMeshComponent* SpotPointer;
	UStaticMeshComponent* RingPointer;
	UStaticMeshComponent* RollingRingPointer;
	UWidgetInteractionComponent* WidgetInteraction;

private:
	void ConfigureResources();
	void ActivatePointer();
	bool IsGazeInteractable();

private:
	UStaticMesh* spotPointerMesh;
	UMaterialInterface* spotPointerMaterial;
	UStaticMesh* ringPointerMesh;
	UMaterialInterface* ringPointerMaterial;
	UStaticMesh* rollingRingPointerMesh;
	UMaterialInterface* rollingRingPointerMaterial;

// Trace
private:
	FVector worldLocation;
	FRotator worldRotation;

	void SetIgnoreActors(FCollisionQueryParams& collisionParams);

	bool tracedPhysics;
	FHitResult hit;
	AActor* focusActor;

	bool tracedWidget;
	FVector tracedWidgetPosition;

	FEyeTrackerGazeData m_GazeData;
	FVector GetEyeVector();
	FRotator GetEyeRotation();

// Trigger type
private:
	bool IsTimeout();
	float gazeOnTime;

	bool IsClicked();
	bool pressState;
};
