// Copyright HTC Corporation. All Rights Reserved.

#include "WaveOpenXRHandInteraction.h"
#include "IXRTrackingSystem.h"
#include "OpenXRCore.h"
#include "UObject/UObjectIterator.h"
#include "IOpenXRExtensionPlugin.h"
#include "Modules/ModuleManager.h"
#include "DrawDebugHelpers.h"
#include "Engine/Engine.h"
#include "WaveOpenXRExtensionPluginDefinition.h"

#if WITH_EDITOR
#include "Editor/EditorEngine.h"
#include "Editor.h"
#endif

DEFINE_LOG_CATEGORY(LogWaveOpenXRHandInteraction);

IMPLEMENT_MODULE(FWaveOpenXRHandInteractionModule, WaveOpenXRHandInteraction);

static TAutoConsoleVariable<int32> CVarEnableOpenXRHandInteractionDebug(TEXT("OpenXR.debug.EnableHandInteractionDebug"), 1, TEXT("0 - HandInteraction debug visualizations are disabled. 1 - HandInteraction debug visualizations are enabled."));

FWaveOpenXRHandInteraction::FWaveOpenXRHandInteraction()
{
	UE_LOG(LogWaveOpenXRHandInteraction, Log, TEXT("FWaveOpenXRHandInteraction() register extension feature."));
	RegisterOpenXRExtensionModularFeature();
}

FWaveOpenXRHandInteraction::~FWaveOpenXRHandInteraction()
{
	Destroy();
}

void FWaveOpenXRHandInteraction::Destroy()
{
}

bool FWaveOpenXRHandInteraction::GetRequiredExtensions(TArray<const ANSICHAR*>& OutExtensions)
{
	UE_LOG(LogWaveOpenXRHandInteraction, Log, TEXT("GetRequiredExtensions() XR_HTC_hand_interaction."));
	OutExtensions.Add("XR_HTC_hand_interaction");
	return true;
}

bool FWaveOpenXRHandInteraction::GetInteractionProfile(XrInstance InInstance, FString& OutKeyPrefix, XrPath& OutPath, bool& OutHasHaptics)
{
	UE_LOG(LogWaveOpenXRHandInteraction, Log, TEXT("GetInteractionProfile() /interaction_profiles/htc/hand_interaction."));
	OutKeyPrefix = "HandInteraction";
	OutHasHaptics = false;
	return xrStringToPath(InInstance, "/interaction_profiles/htc/hand_interaction", &OutPath) == XR_SUCCESS;
}

void FWaveOpenXRHandInteraction::AddActions(XrInstance Instance, TFunction<XrAction(XrActionType InActionType, const FName& InName, const TArray<XrPath>& InSubactionPaths)> AddAction)
{
	// Create suggested bindings
	XrPath HandInteractionProfilePath = XR_NULL_PATH;
	XR_ENSURE(xrStringToPath(Instance, "/interaction_profiles/htc/hand_interaction", &HandInteractionProfilePath));

	TArray<XrPath> subactionPaths;
	XrActionSuggestedBinding bindings[4];

	// 0. left aim pose
	UE_LOG(LogWaveOpenXRHandInteraction, Log, TEXT("AddActions() 0. /user/hand_htc/left/input/aim/pose."));
	LeftAimPoseAction = AddAction(XR_ACTION_TYPE_POSE_INPUT, "left_aim_pose", subactionPaths);
	if (LeftAimPoseAction == XR_NULL_HANDLE) { return; }
	XrPath LeftAimPosePath = XR_NULL_PATH;
	XR_ENSURE(xrStringToPath(Instance, "/user/hand_htc/left/input/aim/pose", &LeftAimPosePath));
	bindings[0].action = LeftAimPoseAction;
	bindings[0].binding = LeftAimPosePath;

	// 1. right aim pose
	UE_LOG(LogWaveOpenXRHandInteraction, Log, TEXT("AddActions() 1. /user/hand_htc/right/input/aim/pose."));
	RightAimPoseAction = AddAction(XR_ACTION_TYPE_POSE_INPUT, "right_aim_pose", subactionPaths);
	if (RightAimPoseAction == XR_NULL_HANDLE) { return; }
	XrPath RightAimPosePath = XR_NULL_PATH;
	XR_ENSURE(xrStringToPath(Instance, "/user/hand_htc/right/input/aim/pose", &RightAimPosePath));
	bindings[1].action = RightAimPoseAction;
	bindings[1].binding = RightAimPosePath;

	// 2. left select value
	UE_LOG(LogWaveOpenXRHandInteraction, Log, TEXT("AddActions() 2. /user/hand_htc/left/input/select/value."));
	LeftSelectAction = AddAction(XR_ACTION_TYPE_FLOAT_INPUT, "left_select_value", subactionPaths);
	if (LeftSelectAction == XR_NULL_HANDLE) { return; }
	XrPath LeftSelectPath = XR_NULL_PATH;
	XR_ENSURE(xrStringToPath(Instance, "/user/hand_htc/left/input/select/value", &LeftSelectPath));
	bindings[2].action = LeftSelectAction;
	bindings[2].binding = LeftSelectPath;

	// 3. right select value
	UE_LOG(LogWaveOpenXRHandInteraction, Log, TEXT("AddActions() 3. /user/hand_htc/right/input/select/value."));
	RightSelectAction = AddAction(XR_ACTION_TYPE_FLOAT_INPUT, "right_select_value", subactionPaths);
	if (RightSelectAction == XR_NULL_HANDLE) { return; }
	XrPath RightSelectPath = XR_NULL_PATH;
	XR_ENSURE(xrStringToPath(Instance, "/user/hand_htc/right/input/select/value", &RightSelectPath));
	bindings[3].action = RightSelectAction;
	bindings[3].binding = RightSelectPath;

	XrInteractionProfileSuggestedBinding SuggestedBindings{ XR_TYPE_INTERACTION_PROFILE_SUGGESTED_BINDING };
	SuggestedBindings.interactionProfile = HandInteractionProfilePath;
	SuggestedBindings.suggestedBindings = bindings;
	SuggestedBindings.countSuggestedBindings = 4;
	XR_ENSURE(xrSuggestInteractionProfileBindings(Instance, &SuggestedBindings));

}

const void* FWaveOpenXRHandInteraction::OnBeginSession(XrSession InSession, const void* InNext)
{
	static FName SystemName(TEXT("OpenXR"));
	if (GEngine->XRSystem.IsValid() && (GEngine->XRSystem->GetSystemName() == SystemName))
	{
		XRTrackingSystem = GEngine->XRSystem.Get();
	}

	// 0. left aim pose, only XR_ACTION_TYPE_POSE_INPUT needs the ActionSpace.
	UE_LOG(LogWaveOpenXRHandInteraction, Log, TEXT("OnBeginSession() xrCreateActionSpace /user/hand_htc/left/input/aim/pose."));
	XrActionSpaceCreateInfo LeftAimPoseActionSpaceInfo{ XR_TYPE_ACTION_SPACE_CREATE_INFO };
	LeftAimPoseActionSpaceInfo.action = LeftAimPoseAction;
	LeftAimPoseActionSpaceInfo.poseInActionSpace = ToXrPose(FTransform::Identity);
	XR_ENSURE(xrCreateActionSpace(InSession, &LeftAimPoseActionSpaceInfo, &LeftAimPoseActionSpace));

	// 1. right aim pose, only XR_ACTION_TYPE_POSE_INPUT needs the ActionSpace.
	UE_LOG(LogWaveOpenXRHandInteraction, Log, TEXT("OnBeginSession() xrCreateActionSpace /user/hand_htc/right/input/aim/pose."));
	XrActionSpaceCreateInfo RightAimPoseActionSpaceInfo{ XR_TYPE_ACTION_SPACE_CREATE_INFO };
	RightAimPoseActionSpaceInfo.action = RightAimPoseAction;
	RightAimPoseActionSpaceInfo.poseInActionSpace = ToXrPose(FTransform::Identity);
	XR_ENSURE(xrCreateActionSpace(InSession, &RightAimPoseActionSpaceInfo, &RightAimPoseActionSpace));

	SyncInfo.countActiveActionSets = 0;
	SyncInfo.activeActionSets = XR_NULL_HANDLE;

	bSessionStarted = true;

	return InNext;
}

void FWaveOpenXRHandInteraction::PostSyncActions(XrSession InSession)
{
	XRTrackingSystem->GetWorldToMetersScale();

	XrActionStateGetInfo LeftAimPoseActionStateInfo{ XR_TYPE_ACTION_STATE_GET_INFO };
	LeftAimPoseActionStateInfo.action = LeftAimPoseAction;
	XR_ENSURE(xrGetActionStatePose(InSession, &LeftAimPoseActionStateInfo, &LeftAimPoseActionState));

	XrActionStateGetInfo RightAimPoseActionStateInfo{ XR_TYPE_ACTION_STATE_GET_INFO };
	RightAimPoseActionStateInfo.action = RightAimPoseAction;
	XR_ENSURE(xrGetActionStatePose(InSession, &RightAimPoseActionStateInfo, &RightAimPoseActionState));

	XrActionStateGetInfo LeftSelectActionStateInfo{ XR_TYPE_ACTION_STATE_GET_INFO };
	LeftSelectActionStateInfo.action = LeftSelectAction;
	XR_ENSURE(xrGetActionStateFloat(InSession, &LeftSelectActionStateInfo, &LeftSelectActionState));

	XrActionStateGetInfo RightSelectActionStateInfo{ XR_TYPE_ACTION_STATE_GET_INFO };
	RightSelectActionStateInfo.action = RightSelectAction;
	XR_ENSURE(xrGetActionStateFloat(InSession, &RightSelectActionStateInfo, &RightSelectActionState));
}

void FWaveOpenXRHandInteraction::UpdateDeviceLocations(XrSession InSession, XrTime DisplayTime, XrSpace TrackingSpace)
{
	HmdOrigin = (uint64_t)TrackingSpace;
	WorldToMetersScale = XRTrackingSystem->GetWorldToMetersScale();

	if (GetPinchActive(true)) {
		XR_ENSURE(xrLocateSpace(LeftAimPoseActionSpace, TrackingSpace, DisplayTime, &LeftAimPoseSpaceLocation));
		if (GetPinchTracked(true))
			PinchTransformLeft = ToFTransform(LeftAimPoseSpaceLocation.pose, WorldToMetersScale);
	}
	if (GetPinchActive(false)) {
		XR_ENSURE(xrLocateSpace(RightAimPoseActionSpace, TrackingSpace, DisplayTime, &RightAimPoseSpaceLocation));
		if (GetPinchTracked(false))
			PinchTransformRight = ToFTransform(RightAimPoseSpaceLocation.pose, WorldToMetersScale);
	}
}

/************************************************************************/
/* FWaveOpenXRHandInteractionModule                                           */
/************************************************************************/

FWaveOpenXRHandInteractionModule::FWaveOpenXRHandInteractionModule()
{
}

void FWaveOpenXRHandInteractionModule::StartupModule()
{
	UE_LOG(LogWaveOpenXRHandInteraction, Log, TEXT("StartupModule()"));
	CHECK_WAVEOPENXR_EXTENSION_FEATURE(wave.HandInteraction.enable);
	IWaveOpenXRHandInteractionModule::StartupModule();
	HandInteraction = TSharedPtr<FWaveOpenXRHandInteraction, ESPMode::ThreadSafe>(new FWaveOpenXRHandInteraction());
}

void FWaveOpenXRHandInteractionModule::ShutdownModule()
{
	UE_LOG(LogWaveOpenXRHandInteraction, Log, TEXT("ShutdownModule()"));
}

TSharedPtr<class FWaveOpenXRHandInteraction, ESPMode::ThreadSafe> FWaveOpenXRHandInteractionModule::CreateHandInteraction()
{
	//UE_LOG(LogWaveOpenXRHandInteraction, Log, TEXT("CreateHandInteraction()"));
	return HandInteraction;
}
