// Copyright HTC Corporation. All Rights Reserved.

#pragma once
#include "IWaveOpenXRHandInteractionModule.h"

#include "CoreMinimal.h"
#include "GameFramework/HUD.h"
#include "Containers/Ticker.h"
#include "IOpenXRExtensionPlugin.h"

#include "OpenXRCore.h"

DECLARE_LOG_CATEGORY_EXTERN(LogWaveOpenXRHandInteraction, Log, All);

class FWaveOpenXRHandInteraction : public IOpenXRExtensionPlugin
{
public:
	FWaveOpenXRHandInteraction();
	virtual ~FWaveOpenXRHandInteraction();
	void Destroy();

	/************************************************************************/
	/* IOpenXRExtensionPlugin                                               */
	/************************************************************************/

	virtual FString GetDisplayName() override
	{
		return FString(TEXT("WaveOpenXRHandInteraction"));
	}
	virtual bool GetRequiredExtensions(TArray<const ANSICHAR*>& OutExtensions) override;
	virtual bool GetInteractionProfile(XrInstance InInstance, FString& OutKeyPrefix, XrPath& OutPath, bool& OutHasHaptics) override;
	virtual void AddActions(XrInstance Instance, TFunction<XrAction(XrActionType InActionType, const FName& InName, const TArray<XrPath>& InSubactionPaths)> AddAction) override;

	virtual const void* OnBeginSession(XrSession InSession, const void* InNext) override;
	virtual void PostSyncActions(XrSession InSession) override;
	virtual void UpdateDeviceLocations(XrSession InSession, XrTime DisplayTime, XrSpace TrackingSpace) override;

private:
	class IXRTrackingSystem* XRTrackingSystem = nullptr;
	float WorldToMetersScale = 100;

	bool bSessionStarted = false;
	XrActionsSyncInfo SyncInfo{ XR_TYPE_ACTIONS_SYNC_INFO };
	XrAction LeftAimPoseAction, RightAimPoseAction, LeftSelectAction, RightSelectAction;
	XrSpace LeftAimPoseActionSpace, RightAimPoseActionSpace;
	XrActionStatePose LeftAimPoseActionState{ XR_TYPE_ACTION_STATE_POSE };
	XrActionStatePose RightAimPoseActionState{ XR_TYPE_ACTION_STATE_POSE };
	XrActionStateFloat LeftSelectActionState{ XR_TYPE_ACTION_STATE_FLOAT };
	XrActionStateFloat RightSelectActionState{ XR_TYPE_ACTION_STATE_FLOAT };

	// HandInteraction cached data
	XrSpaceLocation LeftAimPoseSpaceLocation{ XR_TYPE_SPACE_LOCATION };
	XrSpaceLocation RightAimPoseSpaceLocation{ XR_TYPE_SPACE_LOCATION };

	FTransform PinchTransformLeft = FTransform::Identity;
	FTransform PinchTransformRight = FTransform::Identity;

	uint64_t HmdOrigin = 0;

// Public functions for WaveOpenXRHandInteractionBP
public:
	uint64_t GetHeadOrigin() { return HmdOrigin; }
	float GetPinchStrength(bool isLeft)
	{
		return isLeft ? LeftSelectActionState.currentState : RightSelectActionState.currentState;
	}
	FQuat GetPinchRotation(bool isLeft)
	{
		return isLeft ? PinchTransformLeft.GetRotation() : PinchTransformRight.GetRotation();
	}
	FVector GetPinchPosition(bool isLeft)
	{
		return isLeft ? PinchTransformLeft.GetLocation() : PinchTransformRight.GetLocation();
	}
	bool GetPinchActive(bool isLeft)
	{
		return isLeft ? LeftAimPoseActionState.isActive > 0 : RightAimPoseActionState.isActive > 0;
	}
	bool GetPinchValid(bool isLeft)
	{
		if (!GetPinchActive(isLeft)) { return false; }

		auto locationFlags = isLeft ? LeftAimPoseSpaceLocation.locationFlags : RightAimPoseSpaceLocation.locationFlags;
		return locationFlags & XR_SPACE_LOCATION_ORIENTATION_VALID_BIT &&
			locationFlags & XR_SPACE_LOCATION_POSITION_VALID_BIT;
	}
	bool GetPinchTracked(bool isLeft)
	{
		if (!GetPinchActive(isLeft)) { return false; }

		auto locationFlags = isLeft ? LeftAimPoseSpaceLocation.locationFlags : RightAimPoseSpaceLocation.locationFlags;
		return locationFlags & XR_SPACE_LOCATION_ORIENTATION_TRACKED_BIT &&
			locationFlags & XR_SPACE_LOCATION_POSITION_TRACKED_BIT;
	}
};

class FWaveOpenXRHandInteractionModule : public IWaveOpenXRHandInteractionModule
{
public:
	FWaveOpenXRHandInteractionModule();
	virtual void StartupModule() override;
	virtual void ShutdownModule() override;

	TSharedPtr< class FWaveOpenXRHandInteraction, ESPMode::ThreadSafe > CreateHandInteraction();

private:
	TSharedPtr<FWaveOpenXRHandInteraction, ESPMode::ThreadSafe> HandInteraction = nullptr;
};
