// Copyright HTC Corporation. All Rights Reserved.

#include "WaveOpenXRWristTracker.h"
#include "IXRTrackingSystem.h"
#include "OpenXRCore.h"
#include "UObject/UObjectIterator.h"
#include "IOpenXRExtensionPlugin.h"
#include "Modules/ModuleManager.h"
#include "DrawDebugHelpers.h"
#include "Engine/Engine.h"
#include "WaveOpenXRExtensionPluginDefinition.h"

#if WITH_EDITOR
#include "Editor/EditorEngine.h"
#include "Editor.h"
#endif

DEFINE_LOG_CATEGORY(LogWaveOpenXRWristTracker);

static TAutoConsoleVariable<int32> CVarEnableOpenXRWristTrackerDebug(TEXT("OpenXR.debug.EnableWristTrackerDebug"), 1, TEXT("0 - WristTracker debug visualizations are disabled. 1 - WristTracker debug visualizations are enabled."));

/************************************************************************/
/* FWaveOpenXRWristTracker                                              */
/************************************************************************/

FWaveOpenXRWristTracker* FWaveOpenXRWristTracker::m_Instance = nullptr;

FWaveOpenXRWristTracker::FWaveOpenXRWristTracker(const TSharedRef<FGenericApplicationMessageHandler>& InMessageHandler)
	: MessageHandler(InMessageHandler)
{
	RegisterOpenXRExtensionModularFeature();

	m_Instance = this;
	UE_LOG(LogWaveOpenXRWristTracker, Log, TEXT("FWaveOpenXRWristTracker() register extension feature WristTracker %p."), m_Instance);
}

FWaveOpenXRWristTracker::~FWaveOpenXRWristTracker()
{
	Destroy();
}

void FWaveOpenXRWristTracker::Destroy()
{
}

bool FWaveOpenXRWristTracker::GetRequiredExtensions(TArray<const ANSICHAR*>& OutExtensions)
{
	UE_LOG(LogWaveOpenXRWristTracker, Log, TEXT("GetRequiredExtensions() XR_HTC_vive_wrist_tracker_interaction."));
	OutExtensions.Add("XR_HTC_vive_wrist_tracker_interaction");
	return true;
}

bool FWaveOpenXRWristTracker::GetInteractionProfile(XrInstance InInstance, FString& OutKeyPrefix, XrPath& OutPath, bool& OutHasHaptics)
{
	UE_LOG(LogWaveOpenXRWristTracker, Log, TEXT("GetInteractionProfile() /interaction_profiles/htc/vive_wrist_tracker."));
	OutKeyPrefix = "WristTracker";
	OutHasHaptics = false;
	return xrStringToPath(InInstance, "/interaction_profiles/htc/vive_wrist_tracker", &OutPath) == XR_SUCCESS;
}

void FWaveOpenXRWristTracker::AddActions(XrInstance Instance, TFunction<XrAction(XrActionType InActionType, const FName& InName, const TArray<XrPath>& InSubactionPaths)> AddAction)
{
	// Create suggested bindings
	XrPath WristTrackerProfilePath = XR_NULL_PATH;
	XR_ENSURE(xrStringToPath(Instance, "/interaction_profiles/htc/vive_wrist_tracker", &WristTrackerProfilePath));

	TArray<XrPath> subactionPaths;
	XrActionSuggestedBinding bindings[kActionCount];

	// 0. left entity pose
	int action_index = 0;
	if (action_index < kActionCount)
	{
		UE_LOG(LogWaveOpenXRWristTracker, Log, TEXT("AddActions() %d. /user/wrist_htc/left/input/entity_htc/pose."), action_index);
		LeftEntityPoseAction = AddAction(XR_ACTION_TYPE_POSE_INPUT, "left_entity_pose", subactionPaths);
		if (LeftEntityPoseAction == XR_NULL_HANDLE) { return; }
		XrPath LeftEntityPosePath = XR_NULL_PATH;
		XR_ENSURE(xrStringToPath(Instance, "/user/wrist_htc/left/input/entity_htc/pose", &LeftEntityPosePath));
		bindings[action_index].action = LeftEntityPoseAction;
		bindings[action_index].binding = LeftEntityPosePath;
	}

	// 1. right entity pose
	action_index++;
	if (action_index < kActionCount)
	{
		UE_LOG(LogWaveOpenXRWristTracker, Log, TEXT("AddActions() %d. /user/wrist_htc/right/input/entity_htc/pose."), action_index);
		RightEntityPoseAction = AddAction(XR_ACTION_TYPE_POSE_INPUT, "right_entity_pose", subactionPaths);
		if (RightEntityPoseAction == XR_NULL_HANDLE) { return; }
		XrPath RightEntityPosePath = XR_NULL_PATH;
		XR_ENSURE(xrStringToPath(Instance, "/user/wrist_htc/right/input/entity_htc/pose", &RightEntityPosePath));
		bindings[action_index].action = RightEntityPoseAction;
		bindings[action_index].binding = RightEntityPosePath;
	}

	// 2. left button x click
	action_index++;
	if (action_index < kActionCount)
	{
		UE_LOG(LogWaveOpenXRWristTracker, Log, TEXT("AddActions() %d. /user/wrist_htc/left/input/x/click."), action_index);
		LeftXClickAction = AddAction(XR_ACTION_TYPE_BOOLEAN_INPUT, "left_x_click", subactionPaths);
		if (LeftXClickAction == XR_NULL_HANDLE) { return; }
		XrPath LeftXClickPath = XR_NULL_PATH;
		XR_ENSURE(xrStringToPath(Instance, "/user/wrist_htc/left/input/x/click", &LeftXClickPath));
		bindings[action_index].action = LeftXClickAction;
		bindings[action_index].binding = LeftXClickPath;
	}

	// 3. left button menu click
	action_index++;
	if (action_index < kActionCount)
	{
		UE_LOG(LogWaveOpenXRWristTracker, Log, TEXT("AddActions() %d. /user/wrist_htc/left/input/menu/click."), action_index);
		LeftMenuClickAction = AddAction(XR_ACTION_TYPE_BOOLEAN_INPUT, "left_menu_click", subactionPaths);
		if (LeftMenuClickAction == XR_NULL_HANDLE) { return; }
		XrPath LeftMenuClickPath = XR_NULL_PATH;
		XR_ENSURE(xrStringToPath(Instance, "/user/wrist_htc/left/input/menu/click", &LeftMenuClickPath));
		bindings[action_index].action = LeftMenuClickAction;
		bindings[action_index].binding = LeftMenuClickPath;
	}

	// 4. right button a click
	action_index++;
	if (action_index < kActionCount)
	{
		UE_LOG(LogWaveOpenXRWristTracker, Log, TEXT("AddActions() %d. /user/wrist_htc/right/input/a/click."), action_index);
		RightAClickAction = AddAction(XR_ACTION_TYPE_BOOLEAN_INPUT, "right_a_click", subactionPaths);
		if (RightAClickAction == XR_NULL_HANDLE) { return; }
		XrPath RightAClickPath = XR_NULL_PATH;
		XR_ENSURE(xrStringToPath(Instance, "/user/wrist_htc/right/input/a/click", &RightAClickPath));
		bindings[action_index].action = RightAClickAction;
		bindings[action_index].binding = RightAClickPath;
	}

	// 5. right button system click
	action_index++;
	if (action_index < kActionCount)
	{
		UE_LOG(LogWaveOpenXRWristTracker, Log, TEXT("AddActions() %d. /user/wrist_htc/right/input/system/click."), action_index);
		RightSystemClickAction = AddAction(XR_ACTION_TYPE_BOOLEAN_INPUT, "right_system_click", subactionPaths);
		if (RightSystemClickAction == XR_NULL_HANDLE) { return; }
		XrPath RightSystemClickPath = XR_NULL_PATH;
		XR_ENSURE(xrStringToPath(Instance, "/user/wrist_htc/right/input/system/click", &RightSystemClickPath));
		bindings[action_index].action = RightSystemClickAction;
		bindings[action_index].binding = RightSystemClickPath;
	}

	XrInteractionProfileSuggestedBinding SuggestedBindings{ XR_TYPE_INTERACTION_PROFILE_SUGGESTED_BINDING };
	SuggestedBindings.interactionProfile = WristTrackerProfilePath;
	SuggestedBindings.suggestedBindings = bindings;
	SuggestedBindings.countSuggestedBindings = kActionCount;
	XR_ENSURE(xrSuggestInteractionProfileBindings(Instance, &SuggestedBindings));

}

const void* FWaveOpenXRWristTracker::OnBeginSession(XrSession InSession, const void* InNext)
{
	static FName SystemName(TEXT("OpenXR"));
	if (GEngine->XRSystem.IsValid() && (GEngine->XRSystem->GetSystemName() == SystemName))
	{
		XRTrackingSystem = GEngine->XRSystem.Get();
	}

	// 0. left entity pose, only XR_ACTION_TYPE_POSE_INPUT needs the ActionSpace.
	UE_LOG(LogWaveOpenXRWristTracker, Log, TEXT("OnBeginSession() xrCreateActionSpace /user/wrist_htc/left/input/entity_htc/pose."));
	XrActionSpaceCreateInfo LeftEntityPoseActionSpaceInfo{ XR_TYPE_ACTION_SPACE_CREATE_INFO };
	LeftEntityPoseActionSpaceInfo.action = LeftEntityPoseAction;
	LeftEntityPoseActionSpaceInfo.poseInActionSpace = ToXrPose(FTransform::Identity);
	XR_ENSURE(xrCreateActionSpace(InSession, &LeftEntityPoseActionSpaceInfo, &LeftEntityPoseActionSpace));

	// 1. right entity pose, only XR_ACTION_TYPE_POSE_INPUT needs the ActionSpace.
	UE_LOG(LogWaveOpenXRWristTracker, Log, TEXT("OnBeginSession() xrCreateActionSpace /user/wrist_htc/right/input/entity_htc/pose."));
	XrActionSpaceCreateInfo RightEntityPoseActionSpaceInfo{ XR_TYPE_ACTION_SPACE_CREATE_INFO };
	RightEntityPoseActionSpaceInfo.action = RightEntityPoseAction;
	RightEntityPoseActionSpaceInfo.poseInActionSpace = ToXrPose(FTransform::Identity);
	XR_ENSURE(xrCreateActionSpace(InSession, &RightEntityPoseActionSpaceInfo, &RightEntityPoseActionSpace));

	SyncInfo.countActiveActionSets = 0;
	SyncInfo.activeActionSets = XR_NULL_HANDLE;

	bSessionStarted = true;

	return InNext;
}

void FWaveOpenXRWristTracker::PostSyncActions(XrSession InSession)
{
	XRTrackingSystem->GetWorldToMetersScale();

	// left entity pose
	XrActionStateGetInfo LeftEntityPoseActionStateInfo{ XR_TYPE_ACTION_STATE_GET_INFO };
	LeftEntityPoseActionStateInfo.action = LeftEntityPoseAction;
	XR_ENSURE(xrGetActionStatePose(InSession, &LeftEntityPoseActionStateInfo, &LeftEntityPoseActionState));

	// right entity pose
	XrActionStateGetInfo RightEntityPoseActionStateInfo{ XR_TYPE_ACTION_STATE_GET_INFO };
	RightEntityPoseActionStateInfo.action = RightEntityPoseAction;
	XR_ENSURE(xrGetActionStatePose(InSession, &RightEntityPoseActionStateInfo, &RightEntityPoseActionState));

	// left button x click
	XrActionStateGetInfo LeftXClickActionStateInfo{ XR_TYPE_ACTION_STATE_GET_INFO };
	LeftXClickActionStateInfo.action = LeftXClickAction;
	XR_ENSURE(xrGetActionStateBoolean(InSession, &LeftXClickActionStateInfo, &LeftXClickActionState));

	// left button menu click
	XrActionStateGetInfo LeftMenuActionStateInfo{ XR_TYPE_ACTION_STATE_GET_INFO };
	LeftMenuActionStateInfo.action = LeftMenuClickAction;
	XR_ENSURE(xrGetActionStateBoolean(InSession, &LeftMenuActionStateInfo, &LeftMenuClickActionState));

	// right button a click
	XrActionStateGetInfo RightAActionStateInfo{ XR_TYPE_ACTION_STATE_GET_INFO };
	RightAActionStateInfo.action = RightAClickAction;
	XR_ENSURE(xrGetActionStateBoolean(InSession, &RightAActionStateInfo, &RightAClickActionState));

	// right button system click
	XrActionStateGetInfo RightSystemActionStateInfo{ XR_TYPE_ACTION_STATE_GET_INFO };
	RightSystemActionStateInfo.action = RightSystemClickAction;
	XR_ENSURE(xrGetActionStateBoolean(InSession, &RightSystemActionStateInfo, &RightSystemClickActionState));
}

void FWaveOpenXRWristTracker::UpdateDeviceLocations(XrSession InSession, XrTime DisplayTime, XrSpace TrackingSpace)
{
	HmdOrigin = (uint64_t)TrackingSpace;
	WorldToMetersScale = XRTrackingSystem->GetWorldToMetersScale();

	if (GetTrackerActive(true)) {
		XR_ENSURE(xrLocateSpace(LeftEntityPoseActionSpace, TrackingSpace, DisplayTime, &LeftEntityPoseSpaceLocation));
		if (GetTrackerTracked(true))
			TrackerTransformLeft = ToFTransform(LeftEntityPoseSpaceLocation.pose, WorldToMetersScale);
	}
	if (GetTrackerActive(false)) {
		XR_ENSURE(xrLocateSpace(RightEntityPoseActionSpace, TrackingSpace, DisplayTime, &RightEntityPoseSpaceLocation));
		if (GetTrackerTracked(false))
			TrackerTransformRight = ToFTransform(RightEntityPoseSpaceLocation.pose, WorldToMetersScale);
	}
}

namespace TrackerKeys
{
	/// ---- Left hand only ----
	// "/input/menu/click"
	const FKey Tracker_Left_Menu_Click("Tracker_Left_Menu_Click");
	// "/input/x/click"
	const FKey Tracker_Left_X_Click("Tracker_Left_X_Click");

	/// ---- Right hand only ----
	// "/input/system/click"
	const FKey Tracker_Right_System_Click("Tracker_Right_System_Click");
	// "/input/a/click"
	const FKey Tracker_Right_A_Click("Tracker_Right_A_Click");
}

#pragma region
void FWaveOpenXRWristTracker::SendControllerEvents()
{
	bool s_KeyStateLeft[kTrackerButtonCount] = { false, false };
	bool s_KeyStateRight[kTrackerButtonCount] = { false, false };

	uint8_t menu_id = (uint8_t)TrackerButton::Menu;
	uint8_t primary_id = (uint8_t)TrackerButton::Primary;

	/// Left tracker keys
	// "/user/wrist_htc/left/input/menu/click"
	s_KeyStateLeft[menu_id] = GetTrackerKeyDown(true, TrackerButton::Menu);
	if (s_KeyStateLeftEx[menu_id] != s_KeyStateLeft[menu_id])
	{
		s_KeyStateLeftEx[menu_id] = s_KeyStateLeft[menu_id];

		FName keyName = TrackerKeys::Tracker_Left_Menu_Click.GetFName();
		UE_LOG(LogWaveOpenXRWristTracker, Log, TEXT("SendControllerEvents() %s = %d."),
			*keyName.ToString(), (uint8_t)s_KeyStateLeft[menu_id]);

		if (s_KeyStateLeftEx[menu_id])
			MessageHandler->OnControllerButtonPressed(keyName, 0, false);
		else
			MessageHandler->OnControllerButtonReleased(keyName, 0, false);
	}

	// "/user/wrist_htc/left/input/x/click"
	s_KeyStateLeft[primary_id] = GetTrackerKeyDown(true, TrackerButton::Primary);
	if (s_KeyStateLeftEx[primary_id] != s_KeyStateLeft[primary_id])
	{
		s_KeyStateLeftEx[primary_id] = s_KeyStateLeft[primary_id];

		FName keyName = TrackerKeys::Tracker_Left_X_Click.GetFName();
		UE_LOG(LogWaveOpenXRWristTracker, Log, TEXT("SendControllerEvents() %s = %d."),
			*keyName.ToString(), (uint8_t)s_KeyStateLeft[primary_id]);

		if (s_KeyStateLeftEx[primary_id])
			MessageHandler->OnControllerButtonPressed(keyName, 0, false);
		else
			MessageHandler->OnControllerButtonReleased(keyName, 0, false);
	}

	/// Right tracker keys
	// "/user/wrist_htc/right/input/system/click"
	s_KeyStateRight[menu_id] = GetTrackerKeyDown(false, TrackerButton::Menu);
	if (s_KeyStateRightEx[menu_id] != s_KeyStateRight[menu_id])
	{
		s_KeyStateRightEx[menu_id] = s_KeyStateRight[menu_id];

		FName keyName = TrackerKeys::Tracker_Right_System_Click.GetFName();
		UE_LOG(LogWaveOpenXRWristTracker, Log, TEXT("SendControllerEvents() %s = %d."),
			*keyName.ToString(), (uint8_t)s_KeyStateRight[menu_id]);

		if (s_KeyStateRightEx[menu_id])
			MessageHandler->OnControllerButtonPressed(keyName, 0, false);
		else
			MessageHandler->OnControllerButtonReleased(keyName, 0, false);
	}

	// "/user/wrist_htc/right/input/a/click"
	s_KeyStateRight[primary_id] = GetTrackerKeyDown(false, TrackerButton::Primary);
	if (s_KeyStateRightEx[primary_id] != s_KeyStateRight[primary_id])
	{
		s_KeyStateRightEx[primary_id] = s_KeyStateRight[primary_id];

		FName keyName = TrackerKeys::Tracker_Right_A_Click.GetFName();
		UE_LOG(LogWaveOpenXRWristTracker, Log, TEXT("SendControllerEvents() %s = %d."),
			*keyName.ToString(), (uint8_t)s_KeyStateRight[primary_id]);

		if (s_KeyStateRightEx[primary_id])
			MessageHandler->OnControllerButtonPressed(keyName, 0, false);
		else
			MessageHandler->OnControllerButtonReleased(keyName, 0, false);
	}
}
#pragma endregion IInputDevice overrides

/************************************************************************/
/* FWaveOpenXRWristTrackerModule                                        */
/************************************************************************/

#define LOCTEXT_NAMESPACE "FWaveOpenXRWristTrackerModule"

FName IWaveOpenXRWristTrackerModule::WaveOpenXRWristTrackerModularKeyName = FName(TEXT("WaveOpenXRWristTracker"));

FWaveOpenXRWristTrackerModule::FWaveOpenXRWristTrackerModule()
{
}

void FWaveOpenXRWristTrackerModule::StartupModule()
{
	UE_LOG(LogWaveOpenXRWristTracker, Log, TEXT("StartupModule()"));
	CHECK_WAVEOPENXR_EXTENSION_FEATURE(wave.WristTracker.enable);
	IWaveOpenXRWristTrackerModule::StartupModule();

	TSharedPtr<FGenericApplicationMessageHandler> DummyMessageHandler(new FGenericApplicationMessageHandler());
	CreateInputDevice(DummyMessageHandler.ToSharedRef());

	EKeys::AddMenuCategoryDisplayInfo("Focus3 Tracker", LOCTEXT("Focus3TrackerSubCategory", "Vive Focus3 Tracker"), TEXT("GraphEditor.PadEvent_16x"));

	/// ---- Left hand only ----
	// "/input/menu/click"
	EKeys::AddKey(FKeyDetails(TrackerKeys::Tracker_Left_Menu_Click, LOCTEXT("Tracker_Left_Menu_Click", "Tracker (L) Menu"), FKeyDetails::GamepadKey | FKeyDetails::NotBlueprintBindableKey, "Focus3 Tracker"));
	// "/input/x/click"
	EKeys::AddKey(FKeyDetails(TrackerKeys::Tracker_Left_X_Click, LOCTEXT("Tracker_Left_X_Click", "Tracker (L) X Press"), FKeyDetails::GamepadKey | FKeyDetails::NotBlueprintBindableKey, "Focus3 Tracker"));

	/// ---- Right hand only ----
	// "/input/system/click"
	EKeys::AddKey(FKeyDetails(TrackerKeys::Tracker_Right_System_Click, LOCTEXT("Tracker_Right_System_Click", "Tracker (R) System"), FKeyDetails::GamepadKey | FKeyDetails::NotBlueprintBindableKey, "Focus3 Tracker"));
	// "/input/a/click"
	EKeys::AddKey(FKeyDetails(TrackerKeys::Tracker_Right_A_Click, LOCTEXT("Tracker_Right_A_Click", "Tracker (R) A Press"), FKeyDetails::GamepadKey | FKeyDetails::NotBlueprintBindableKey, "Focus3 Tracker"));
}

void FWaveOpenXRWristTrackerModule::ShutdownModule()
{
	UE_LOG(LogWaveOpenXRWristTracker, Log, TEXT("ShutdownModule()"));
}

TSharedPtr<class IInputDevice> FWaveOpenXRWristTrackerModule::CreateInputDevice(const TSharedRef<FGenericApplicationMessageHandler>& InMessageHandler)
{
	if (!WristTracker.IsValid())
	{
		auto InputDevice = new FWaveOpenXRWristTracker(InMessageHandler);
		WristTracker = TSharedPtr<FWaveOpenXRWristTracker>(InputDevice);
		UE_LOG(LogWaveOpenXRWristTracker, Log, TEXT("CreateInputDevice() new WristTracker %p"), WristTracker.Get());

		return WristTracker;
	}
	else
	{
		WristTracker.Get()->SetMessageHandler(InMessageHandler);
		UE_LOG(LogWaveOpenXRWristTracker, Log, TEXT("CreateInputDevice() update WristTracker %p"), WristTracker.Get());

		return WristTracker;
	}
	return nullptr;
}

FWaveOpenXRWristTracker* FWaveOpenXRWristTrackerModule::GetWristTracker()
{
	return FWaveOpenXRWristTracker::GetInstance();
}

IMPLEMENT_MODULE(FWaveOpenXRWristTrackerModule, WaveOpenXRWristTracker);
