// Copyright HTC Corporation. All Rights Reserved.

#pragma once
#include "IWaveOpenXRWristTrackerModule.h"
#include "WaveOpenXRTrackerEnums.h"

#include "CoreMinimal.h"
#include "GameFramework/HUD.h"
#include "Containers/Ticker.h"
#include "IOpenXRExtensionPlugin.h"
#include "IInputDevice.h"

#include "OpenXRCore.h"

DECLARE_LOG_CATEGORY_EXTERN(LogWaveOpenXRWristTracker, Log, All);

class FWaveOpenXRWristTracker : public IOpenXRExtensionPlugin, public IInputDevice
{
public:
	FWaveOpenXRWristTracker(const TSharedRef<FGenericApplicationMessageHandler>& InMessageHandler);
	virtual ~FWaveOpenXRWristTracker();
	void Destroy();

	/************************************************************************/
	/* IOpenXRExtensionPlugin                                               */
	/************************************************************************/

	virtual FString GetDisplayName() override
	{
		return FString(TEXT("WaveOpenXRWristTracker"));
	}
	virtual bool GetRequiredExtensions(TArray<const ANSICHAR*>& OutExtensions) override;
	virtual bool GetInteractionProfile(XrInstance InInstance, FString& OutKeyPrefix, XrPath& OutPath, bool& OutHasHaptics) override;
	virtual void AddActions(XrInstance Instance, TFunction<XrAction(XrActionType InActionType, const FName& InName, const TArray<XrPath>& InSubactionPaths)> AddAction) override;

	virtual const void* OnBeginSession(XrSession InSession, const void* InNext) override;
	virtual void PostSyncActions(XrSession InSession) override;
	virtual void UpdateDeviceLocations(XrSession InSession, XrTime DisplayTime, XrSpace TrackingSpace) override;

	static inline FWaveOpenXRWristTracker* GetInstance() {
		return m_Instance;
	}
protected:
	static FWaveOpenXRWristTracker* m_Instance;

private:
	class IXRTrackingSystem* XRTrackingSystem = nullptr;
	float WorldToMetersScale = 100;

	bool bSessionStarted = false;
	XrActionsSyncInfo SyncInfo{ XR_TYPE_ACTIONS_SYNC_INFO };
	static const int kActionCount = 6;

	// Pose
	XrAction LeftEntityPoseAction, RightEntityPoseAction;
	XrSpace LeftEntityPoseActionSpace, RightEntityPoseActionSpace;
	XrActionStatePose LeftEntityPoseActionState{ XR_TYPE_ACTION_STATE_POSE };
	XrActionStatePose RightEntityPoseActionState{ XR_TYPE_ACTION_STATE_POSE };
	// Left button
	XrAction LeftXClickAction, LeftMenuClickAction;
	// Right buttons
	XrAction RightAClickAction, RightSystemClickAction;

	// Pose cached data
	XrSpaceLocation LeftEntityPoseSpaceLocation{ XR_TYPE_SPACE_LOCATION };
	XrSpaceLocation RightEntityPoseSpaceLocation{ XR_TYPE_SPACE_LOCATION };
	FTransform TrackerTransformLeft = FTransform::Identity;
	FTransform TrackerTransformRight = FTransform::Identity;
	// Left button cached data
	XrActionStateBoolean LeftXClickActionState{ XR_TYPE_ACTION_STATE_BOOLEAN };
	XrActionStateBoolean LeftMenuClickActionState{ XR_TYPE_ACTION_STATE_BOOLEAN };
	// Right button cached data
	XrActionStateBoolean RightAClickActionState{ XR_TYPE_ACTION_STATE_BOOLEAN };
	XrActionStateBoolean RightSystemClickActionState{ XR_TYPE_ACTION_STATE_BOOLEAN };

	uint64_t HmdOrigin = 0;

// Public functions for WaveOpenXRWristTrackerBP
public:
	uint64_t GetHeadOrigin() { return HmdOrigin; }
	FQuat GetTrackerRotation(bool isLeft)
	{
		auto rotation = isLeft ? TrackerTransformLeft.GetRotation() : TrackerTransformRight.GetRotation();
		return rotation;
	}
	FVector GetTrackerPosition(bool isLeft)
	{
		auto location = isLeft ? TrackerTransformLeft.GetLocation() : TrackerTransformRight.GetLocation();
		return location;
	}
	bool GetTrackerActive(bool isLeft)
	{
		return isLeft ? LeftEntityPoseActionState.isActive > 0 : RightEntityPoseActionState.isActive > 0;
	}
	bool GetTrackerValid(bool isLeft)
	{
		if (!GetTrackerActive(isLeft)) { return false; }

		auto locationFlags = isLeft ? LeftEntityPoseSpaceLocation.locationFlags : RightEntityPoseSpaceLocation.locationFlags;
		return locationFlags & XR_SPACE_LOCATION_ORIENTATION_VALID_BIT &&
			locationFlags & XR_SPACE_LOCATION_POSITION_VALID_BIT;
	}
	bool GetTrackerTracked(bool isLeft)
	{
		if (!GetTrackerActive(isLeft)) { return false; }

		auto locationFlags = isLeft ? LeftEntityPoseSpaceLocation.locationFlags : RightEntityPoseSpaceLocation.locationFlags;
		return locationFlags & XR_SPACE_LOCATION_ORIENTATION_TRACKED_BIT &&
			locationFlags & XR_SPACE_LOCATION_POSITION_TRACKED_BIT;
	}
	bool GetTrackerKeyDown(bool isLeft, TrackerButton key)
	{
		switch (key)
		{
		case TrackerButton::Menu:
			return isLeft ? (LeftMenuClickActionState.currentState > 0) : (RightSystemClickActionState.currentState > 0);
		case TrackerButton::Primary:
			return isLeft ? (LeftXClickActionState.currentState > 0) : (RightAClickActionState.currentState > 0);
		default:
			break;
		}
		return false;
	}

public:

	/************************************************************************/
	/* IInputDevice                                                         */
	/************************************************************************/

	virtual bool Exec(UWorld* InWorld, const TCHAR* Cmd, FOutputDevice& Ar) override { return false; }
	virtual void SetMessageHandler(const TSharedRef< FGenericApplicationMessageHandler >& InMessageHandler) override
	{
		MessageHandler = InMessageHandler;
	}
	virtual void SendControllerEvents() override;

	virtual void SetChannelValue(int32 ControllerId, FForceFeedbackChannelType ChannelType, float Value) override {}
	virtual void SetChannelValues(int32 ControllerId, const FForceFeedbackValues& Values) override {}
	virtual void Tick(float DeltaTime) override {}
private:
	/** The recipient of motion controller input events */
	TSharedPtr< FGenericApplicationMessageHandler > MessageHandler;
	bool s_KeyStateLeftEx[kTrackerButtonCount] = { false, false };
	bool s_KeyStateRightEx[kTrackerButtonCount] = { false, false };
};

class FWaveOpenXRWristTrackerModule : public IWaveOpenXRWristTrackerModule
{
public:
	FWaveOpenXRWristTrackerModule();
	virtual void StartupModule() override;
	virtual void ShutdownModule() override;
	virtual TSharedPtr<class IInputDevice> CreateInputDevice(const TSharedRef<FGenericApplicationMessageHandler>& InMessageHandler) override;

	FWaveOpenXRWristTracker* GetWristTracker();

private:
	TSharedPtr<FWaveOpenXRWristTracker> WristTracker;
};
